==> netexpect_0.20.orig.tar.bz2: netexpect-0.20/src/packets/proto_hier_stats.c <==
/* proto_hier_stats.c
 * Routines for calculating statistics based on protocol.
 *
 * $Id: proto_hier_stats.c 23786 2007-12-06 18:10:35Z stig $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif

#include <epan/epan_dissect.h>

#include <string.h>
#include <glib.h>
#include <tcl.h>

#include "proto_hier_stats.h"
#include "tcl_packet.h"

#define STAT_NODE_STATS(n)   ((ph_stats_node_t*)(n)->data)
#define STAT_NODE_HFINFO(n)  (STAT_NODE_STATS(n)->hfinfo)


static GNode*
find_stat_node(GNode *parent_stat_node, header_field_info *needle_hfinfo)
{
    GNode *needle_stat_node;
    header_field_info *hfinfo;
    ph_stats_node_t *stats;

    for (needle_stat_node = g_node_first_child(parent_stat_node);
	 needle_stat_node;
	 needle_stat_node = g_node_next_sibling(needle_stat_node) ) {
	hfinfo = STAT_NODE_HFINFO(needle_stat_node);
	if (hfinfo &&  hfinfo->id == needle_hfinfo->id)
	    return needle_stat_node;
    }

    /* None found. Create one. */
    stats = g_new(ph_stats_node_t, 1);

    /* Intialize counters */
    stats->hfinfo = needle_hfinfo;
    stats->num_pkts_total = 0;
    stats->num_pkts_last = 0;
    stats->num_bytes_total = 0;
    stats->num_bytes_last = 0;

    needle_stat_node = g_node_new(stats);
    g_node_append(parent_stat_node, needle_stat_node);
    return needle_stat_node;
}


static void
process_node(proto_node *ptree_node, GNode *parent_stat_node, ph_stats_t *ps, guint pkt_len)
{
	field_info		*finfo;
	ph_stats_node_t		*stats;
	proto_node		*proto_sibling_node;
	GNode			*stat_node;

	finfo = PITEM_FINFO(ptree_node);
	g_assert(finfo);

	/* If the field info isn't related to a protocol but to a field,
	 * don't count them, as they don't belong to any protocol.
	 * (happens e.g. for toplevel tree item of desegmentation "[Reassembled TCP Segments]") */
	if (finfo->hfinfo->parent != -1) {
		/* Skip this element, use parent status node */
		stat_node = parent_stat_node;
		stats = STAT_NODE_STATS(stat_node);
	} else {
		stat_node = find_stat_node(parent_stat_node, finfo->hfinfo);

		stats = STAT_NODE_STATS(stat_node);
		stats->num_pkts_total++;
		stats->num_bytes_total += pkt_len;
	}

	proto_sibling_node = ptree_node->next;

	if (proto_sibling_node) {
		/* If the name does not exist for this proto_sibling_node, then it is
		 * not a normal protocol in the top-level tree.  It was instead
		 * added as a normal tree such as IPv6's Hop-by-hop Option Header and
		 * should be skipped when creating the protocol hierarchy display. */
		if (strlen(proto_sibling_node->finfo->hfinfo->name) == 0)
			proto_sibling_node = proto_sibling_node->next;

		process_node(proto_sibling_node, stat_node, ps, pkt_len);
	} else {
		stats->num_pkts_last++;
		stats->num_bytes_last += pkt_len;
	}
}

static void
process_tree(proto_tree *protocol_tree, ph_stats_t* ps, guint pkt_len)
{
	proto_node	*ptree_node;

	ptree_node = ((proto_node *)protocol_tree)->first_child;
	if (!ptree_node) {
		return;
	}

	process_node(ptree_node, ps->stats_tree, ps, pkt_len);
}

static gboolean
process_frame(frame_data *frame, uint8_t *pkt, ph_stats_t* ps)
{
	epan_dissect_t			*edt;
	union wtap_pseudo_header	phdr;
	double				cur_time;

	memset(&phdr, 0, sizeof(phdr) );

	/* Dissect the frame */
	edt = epan_dissect_new(TRUE /* tree */, FALSE /* not visible */);
	epan_dissect_run(edt, &phdr, pkt, frame, NULL);

	/* Get stats from this protocol tree */
	process_tree(edt->tree, ps, frame->pkt_len);

	/* Update times */
	cur_time = nstime_to_sec(&frame->abs_ts);
	if (cur_time < ps->first_time) {
	  ps->first_time = cur_time;
	}
	if (cur_time > ps->last_time){
	  ps->last_time = cur_time;
	}

	/* Free our memory. */
	epan_dissect_free(edt);

	return TRUE;	/* success */
}

ph_stats_t *
ph_stats_tcl(Tcl_Interp *interp, int objc, Tcl_Obj * const *objv)
{
    ph_stats_t *ps;
    frame_data frame;
    guint tot_packets, tot_bytes;
    gboolean stop_flag;
    double cur_time;
    GTimeVal start_time;
    Tcl_ObjType *listobjtype;
    Tcl_Obj *obj, **pkt_list;
    uint8_t *pkt;
    int i, j, llen;

    /* Initialize the data */
    ps = g_new(ph_stats_t, 1);
    ps->tot_packets = 0;
    ps->tot_bytes = 0;
    ps->stats_tree = g_node_new(NULL);
    ps->first_time = 0.0;
    ps->last_time = 0.0;

    stop_flag = FALSE;
    g_get_current_time(&start_time);

    tot_packets = 0;
    tot_bytes = 0;

    listobjtype = Tcl_GetObjType("list");

    for (i = 1; i < objc; i++) {
	obj = Tcl_ObjGetVar2(interp, objv[i], NULL, 0);
	if (!obj)
	    /*
	     * Silently ignore errors. Most likely error here is
	     * inexistant variable.
	     */
	    continue;

	if (obj->typePtr == listobjtype) {
	    Tcl_ListObjGetElements(NULL, obj, &llen, &pkt_list);
	} else {
	    pkt_list = NULL;
	    llen = 1;
	}

	for (j = 0; j < llen; j++) {
	    if (pkt_list)
		/*
		 * Retrieve list element. If not dealing with a list then
		 * the packet object is already in the "obj" variable.
		 */
		obj = pkt_list[j];

	    if (Tcl_GetPacketFromObj(obj, &frame, &pkt) == -1)
		continue;

	    if (tot_packets == 0) {
		cur_time = nstime_to_sec(&frame.abs_ts);
		ps->first_time = cur_time;
		ps->last_time = cur_time;
	    }
	    
	    if (!process_frame(&frame, pkt, ps)) {
		/*
		 * Give up, and set "stop_flag" so we
		 * just abort rather than popping up
		 * the statistics window.
		 */
		stop_flag = TRUE;
		goto stop;
	    }

	    tot_packets++;
	    tot_bytes += frame.pkt_len;
	}
    }

stop:
    if (stop_flag) {
	/*
	 * We quit in the middle; throw away the statistics
	 * and return NULL, so our caller doesn't do anything
	 * with the incomplete statistics.
	 */
	ph_stats_free(ps);
	return NULL;
    }

    ps->tot_packets = tot_packets;
    ps->tot_bytes = tot_bytes;

    return ps;
}

static gboolean
stat_node_free(GNode *node, gpointer data _U_)
{
	ph_stats_node_t	*stats = node->data;

	if (stats) {
		g_free(stats);
	}
	return FALSE;
}

void
ph_stats_free(ph_stats_t *ps)
{

	if (ps->stats_tree) {
		g_node_traverse(ps->stats_tree, G_IN_ORDER,
				G_TRAVERSE_ALL, -1,
				stat_node_free, NULL);
		g_node_destroy(ps->stats_tree);
	}

	g_free(ps);
}

==> netexpect_0.21.orig.tar.bz2: netexpect-0.21/src/packets/proto_hier_stats.c <==
/* proto_hier_stats.c
 * Routines for calculating statistics based on protocol.
 *
 * $Id: proto_hier_stats.c 23786 2007-12-06 18:10:35Z stig $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif

#include <epan/epan_dissect.h>

#include <string.h>
#include <glib.h>
#include <tcl.h>

#include "proto_hier_stats.h"
#include "tcl_packet.h"

#define STAT_NODE_STATS(n)   ((ph_stats_node_t*)(n)->data)
#define STAT_NODE_HFINFO(n)  (STAT_NODE_STATS(n)->hfinfo)


static GNode*
find_stat_node(GNode *parent_stat_node, header_field_info *needle_hfinfo)
{
    GNode *needle_stat_node;
    header_field_info *hfinfo;
    ph_stats_node_t *stats;

    for (needle_stat_node = g_node_first_child(parent_stat_node);
	 needle_stat_node;
	 needle_stat_node = g_node_next_sibling(needle_stat_node) ) {
	hfinfo = STAT_NODE_HFINFO(needle_stat_node);
	if (hfinfo &&  hfinfo->id == needle_hfinfo->id)
	    return needle_stat_node;
    }

    /* None found. Create one. */
    stats = g_new(ph_stats_node_t, 1);

    /* Intialize counters */
    stats->hfinfo = needle_hfinfo;
    stats->num_pkts_total = 0;
    stats->num_pkts_last = 0;
    stats->num_bytes_total = 0;
    stats->num_bytes_last = 0;

    needle_stat_node = g_node_new(stats);
    g_node_append(parent_stat_node, needle_stat_node);
    return needle_stat_node;
}


static void
process_node(proto_node *ptree_node, GNode *parent_stat_node, ph_stats_t *ps, guint pkt_len)
{
	field_info		*finfo;
	ph_stats_node_t		*stats;
	proto_node		*proto_sibling_node;
	GNode			*stat_node;

	finfo = PITEM_FINFO(ptree_node);
	g_assert(finfo);

	/* If the field info isn't related to a protocol but to a field,
	 * don't count them, as they don't belong to any protocol.
	 * (happens e.g. for toplevel tree item of desegmentation "[Reassembled TCP Segments]") */
	if (finfo->hfinfo->parent != -1) {
		/* Skip this element, use parent status node */
		stat_node = parent_stat_node;
		stats = STAT_NODE_STATS(stat_node);
	} else {
		stat_node = find_stat_node(parent_stat_node, finfo->hfinfo);

		stats = STAT_NODE_STATS(stat_node);
		stats->num_pkts_total++;
		stats->num_bytes_total += pkt_len;
	}

	proto_sibling_node = ptree_node->next;

	if (proto_sibling_node) {
		/* If the name does not exist for this proto_sibling_node, then it is
		 * not a normal protocol in the top-level tree.  It was instead
		 * added as a normal tree such as IPv6's Hop-by-hop Option Header and
		 * should be skipped when creating the protocol hierarchy display. */
		if (strlen(proto_sibling_node->finfo->hfinfo->name) == 0)
			proto_sibling_node = proto_sibling_node->next;

		process_node(proto_sibling_node, stat_node, ps, pkt_len);
	} else {
		stats->num_pkts_last++;
		stats->num_bytes_last += pkt_len;
	}
}

static void
process_tree(proto_tree *protocol_tree, ph_stats_t* ps, guint pkt_len)
{
	proto_node	*ptree_node;

	ptree_node = ((proto_node *)protocol_tree)->first_child;
	if (!ptree_node) {
		return;
	}

	process_node(ptree_node, ps->stats_tree, ps, pkt_len);
}

static gboolean
process_frame(struct wtap_pkthdr *phdr, frame_data *frame, uint8_t *pkt,
	      ph_stats_t* ps)
{
	epan_dissect_t			*edt;
	double				cur_time;

	/* Dissect the frame */
	edt = epan_dissect_new(TRUE /* tree */, FALSE /* not visible */);
	epan_dissect_run(edt, phdr, pkt, frame, NULL);

	/* Get stats from this protocol tree */
	process_tree(edt->tree, ps, frame->pkt_len);

	/* Update times */
	cur_time = nstime_to_sec(&frame->abs_ts);
	if (cur_time < ps->first_time) {
	  ps->first_time = cur_time;
	}
	if (cur_time > ps->last_time){
	  ps->last_time = cur_time;
	}

	/* Free our memory. */
	epan_dissect_free(edt);

	return TRUE;	/* success */
}

ph_stats_t *
ph_stats_tcl(Tcl_Interp *interp, int objc, Tcl_Obj * const *objv)
{
    ph_stats_t *ps;
    struct wtap_pkthdr phdr;
    frame_data frame;
    guint tot_packets, tot_bytes;
    gboolean stop_flag;
    double cur_time;
    GTimeVal start_time;
    Tcl_ObjType *listobjtype;
    Tcl_Obj *obj, **pkt_list;
    uint8_t *pkt;
    int i, j, llen;

    /* Initialize the data */
    ps = g_new(ph_stats_t, 1);
    ps->tot_packets = 0;
    ps->tot_bytes = 0;
    ps->stats_tree = g_node_new(NULL);
    ps->first_time = 0.0;
    ps->last_time = 0.0;

    stop_flag = FALSE;
    g_get_current_time(&start_time);

    tot_packets = 0;
    tot_bytes = 0;

    listobjtype = Tcl_GetObjType("list");

    for (i = 1; i < objc; i++) {
	obj = Tcl_ObjGetVar2(interp, objv[i], NULL, 0);
	if (!obj)
	    /*
	     * Silently ignore errors. Most likely error here is
	     * inexistant variable.
	     */
	    continue;

	if (obj->typePtr == listobjtype) {
	    Tcl_ListObjGetElements(NULL, obj, &llen, &pkt_list);
	} else {
	    pkt_list = NULL;
	    llen = 1;
	}

	for (j = 0; j < llen; j++) {
	    if (pkt_list)
		/*
		 * Retrieve list element. If not dealing with a list then
		 * the packet object is already in the "obj" variable.
		 */
		obj = pkt_list[j];

	    if (Tcl_GetPacketFromObj(obj, &phdr, &frame, &pkt) == -1)
		continue;

	    if (tot_packets == 0) {
		cur_time = nstime_to_sec(&frame.abs_ts);
		ps->first_time = cur_time;
		ps->last_time = cur_time;
	    }
	    
	    if (!process_frame(&phdr, &frame, pkt, ps)) {
		/*
		 * Give up, and set "stop_flag" so we
		 * just abort rather than popping up
		 * the statistics window.
		 */
		stop_flag = TRUE;
		goto stop;
	    }

	    tot_packets++;
	    tot_bytes += frame.pkt_len;
	}
    }

stop:
    if (stop_flag) {
	/*
	 * We quit in the middle; throw away the statistics
	 * and return NULL, so our caller doesn't do anything
	 * with the incomplete statistics.
	 */
	ph_stats_free(ps);
	return NULL;
    }

    ps->tot_packets = tot_packets;
    ps->tot_bytes = tot_bytes;

    return ps;
}

static gboolean
stat_node_free(GNode *node, gpointer data _U_)
{
	ph_stats_node_t	*stats = node->data;

	if (stats) {
		g_free(stats);
	}
	return FALSE;
}

void
ph_stats_free(ph_stats_t *ps)
{

	if (ps->stats_tree) {
		g_node_traverse(ps->stats_tree, G_IN_ORDER,
				G_TRAVERSE_ALL, -1,
				stat_node_free, NULL);
		g_node_destroy(ps->stats_tree);
	}

	g_free(ps);
}
